__version__ = '0.5.1'
__author__ = 'John Connell - The Mathematic Fanatic'
__description__ = 'Manim plugin which aims to make it much easier to meaningfully transform algebra expressions.'

from .expressions import *
from .actions import *
from .timelines import *
